# SPDX-FileCopyrightText: 2020 jack-mil
#
# SPDX-License-Identifier: MIT

from bing_rewards.app import main

# __main__ module enabled running as `python -m bing_rewards`
# This is not recommended

main()
